/* This file contains all the functions required for the standalone
   ip_conntrack module.

   These are not required by the compatibility layer.
*/

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */

#define EXPORT_SYMTAB
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>

#include "conntrack/struct_reserve.h"
#include "conntrack/ip_conntrack.h"
#include "conntrack/ip_conntrack_protocol.h"
#include "conntrack/ip_conntrack_core.h"
#include "conntrack/ip_conntrack_helper.h"
#include "common/listhelp.h"

struct module *ip_conntrack_module = &__this_module;

static unsigned int 
print_tuple(char *buffer, const struct ip_conntrack_tuple *tuple,
	    struct ip_conntrack_protocol *proto)
{
	int len;

	len = sprintf(buffer, "src=%u.%u.%u.%u dst=%u.%u.%u.%u ",
		      IP_PARTS(tuple->src.ip), IP_PARTS(tuple->dst.ip));

	len += proto->print_tuple(buffer + len, tuple);

	return len;
}

/* FIXME: Don't print source proto part. --RR */
static unsigned int 
print_expect(char *buffer, const struct ip_conntrack_expect *expect)
{
	unsigned int len;

	len = sprintf(buffer, "EXPECTING: proto=%u ",
		      expect->tuple.dst.protonum);
	len += print_tuple(buffer + len, &expect->tuple,
			   find_proto(expect->tuple.dst.protonum));
	len += sprintf(buffer + len, "\n");
	return len;
}

static unsigned int 
print_conntrack(char *buffer, const struct ip_conntrack *conntrack)
{
	unsigned int len;
	struct ip_conntrack_protocol *proto
		= find_proto(conntrack->tuplehash[0].tuple.dst.protonum);

	len = sprintf(buffer, "%-8s %u %lu ", 
		      proto->name, conntrack->tuplehash[0].tuple.dst.protonum,
		      timer_pending(&conntrack->timeout) 
		      ? (conntrack->timeout.expires - jiffies)/HZ : 0);

	len += proto->print_conntrack(buffer + len, conntrack);
	len += print_tuple(buffer + len, 
			   &conntrack->tuplehash[0].tuple, proto);
	if (conntrack->status == IPS_AWAITING_REPLY 
	    || conntrack->status == IPS_EXPECTED_AWAITING_REPLY)
		len += sprintf(buffer + len, "[UNREPLIED] ");
	len += print_tuple(buffer + len, 
			   &conntrack->tuplehash[1].tuple, proto);
	len += sprintf(buffer + len, "\n");

	return len;
}

/* Returns true when finished. */
static int
conntrack_iterate(const struct ip_conntrack_tuple_hash *hash,
		  char *buffer, off_t offset, off_t *upto,
		  unsigned int *len, unsigned int maxlen)
{
	unsigned int newlen;
	IP_CONNTRACK_ASSERT(hash->ctrack);

	/* Don't print if both halves dead. */
	/* Jul 23 1999: At the moment, this won't happen.  May later. --RR */
	if (hash->dead && hash->ctrack->tuplehash[!DIRECTION(hash)].dead)
		return 0;

	/* Only count originals */
	if (DIRECTION(hash))
		return 0;

	if ((*upto)++ < offset) 
		return 0;

	newlen = print_conntrack(buffer + *len, hash->ctrack);
	if (*len + newlen > maxlen)
		return 1;
	else *len += newlen;

	return 0;
}

static int 
list_conntracks(char *buffer, char **start, off_t offset, int length,
		int dummy)
{
	unsigned int i;
	unsigned int len = 0;
	off_t upto = 0;
	unsigned long flags;
	struct list_head *e;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);

	/* Traverse hash; print originals then reply. */
	for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++) {
		if (LIST_FIND(&ip_conntrack_hash[i], conntrack_iterate,
			      struct ip_conntrack_tuple_hash *,
			      buffer, offset, &upto, &len, length))
			goto finished;
	}

	/* Now iterate through expecteds. */
	for (e = expect_list.next; e != &expect_list; e = e->next) {
		unsigned int last_len;
		struct ip_conntrack_expect *expect 
			= (struct ip_conntrack_expect *)e;
		if (upto++ < offset) continue;

		last_len = len;
		len += print_expect(buffer + len, expect);
		if (len > length) {
			len = last_len;
			goto finished;
		}
	}

 finished:
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);

	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}

/* Connection tracking may drop packets, but never alters them, so
   make it the first hook. */
static struct nf_hook_ops ip_conntrack_in_ops 
= { { NULL, NULL }, ip_conntrack_in, NULL, PF_INET, NF_IP_PRE_ROUTING, -200 };
static struct nf_hook_ops ip_conntrack_local_out_ops 
= { { NULL, NULL }, ip_conntrack_in, NULL, PF_INET, NF_IP_LOCAL_OUT, -200 };

#ifdef CONFIG_PROC_FS
struct proc_dir_entry ip_conntrack_procinfo = {
	0, 12, "ip_conntrack", S_IFREG | S_IRUGO,
	1, 0, 0, 0, &proc_net_inode_operations, &list_conntracks
};
#endif

static int init_or_cleanup(int init)
{
	int ret = 0;

	if (!init) goto cleanup;

	ip_conntrack_init();

#ifdef CONFIG_PROC_FS
	proc_net_register(&ip_conntrack_procinfo);
#endif
	ret = nf_register_hook(&ip_conntrack_in_ops);
	if (ret < 0) {
		printk("ip_conntrack_init: can't register in hook.\n");
		goto cleanup_init;
	}
	ret = nf_register_hook(&ip_conntrack_local_out_ops);
	if (ret < 0) {
		printk("ip_conntrack_init: can't register local out hook.\n");
		goto cleanup_inops;
	}

	return ret;

 cleanup:
	nf_unregister_hook(&ip_conntrack_local_out_ops);
	
 cleanup_inops:
	nf_unregister_hook(&ip_conntrack_in_ops);
 cleanup_init:
#ifdef CONFIG_PROC_FS
	proc_net_unregister(ip_conntrack_procinfo.low_ino);
#endif
	ip_conntrack_cleanup();
	return ret;
}

/* FIXME: Allow NULL functions and sub in pointers to generic for
   them. --RR */
int ip_conntrack_register_protocol(struct ip_conntrack_protocol *proto)
{
	MOD_INC_USE_COUNT;
	/* Not unknown protocol? */
	if (find_proto(proto->proto)->proto != 0)
		return 0;
	list_prepend(&protocol_list, proto);
	return 1;
}

void ip_conntrack_unregister_protocol(struct ip_conntrack_protocol *proto)
{
	unsigned int i;
	unsigned long flags;
	
	LIST_DELETE(&protocol_list, proto);

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);

	/* Delete those using this proto, if they are doing so specifically. */
	/* Traverse hash. */
	for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++) {
		struct list_head *h;

		for (h = ip_conntrack_hash[i].next;
		     h != &ip_conntrack_hash[i]; ) {
			struct ip_conntrack_tuple_hash *hash = (void *)h;
			struct list_head *next = h->next;

			if (hash->tuple.dst.protonum == proto->proto)
				ip_conntrack_delete(hash->ctrack);

			h = next;
		}
	}
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
	MOD_DEC_USE_COUNT;
}

int init_module(void)
{
	return init_or_cleanup(1);
}

void cleanup_module(void)
{
	init_or_cleanup(0);
}

EXPORT_SYMBOL(ip_conntrack_register_protocol);
EXPORT_SYMBOL(ip_conntrack_unregister_protocol);
EXPORT_SYMBOL(expect_list);
EXPORT_SYMBOL(invert_tuplepr);
EXPORT_SYMBOL(ip_conntrack_alter_reply);
EXPORT_SYMBOL(ip_conntrack_destroyed);
EXPORT_SYMBOL(ip_conntrack_find);
EXPORT_SYMBOL(ip_conntrack_module);
EXPORT_STRUCT_RESERVABLE(ip_conntrack);
EXPORT_SYMBOL(ip_conntrack_helper_register);
EXPORT_SYMBOL(ip_conntrack_helper_unregister);
EXPORT_SYMBOL(ip_conntrack_lck);
EXPORT_SYMBOL(ip_conntrack_findinfo);
EXPORT_SYMBOL(ip_conntrack_hash);
EXPORT_SYMBOL(ip_conntrack_delete);
