/*
 * This is a module which is used for logging packets.
 */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/icmp.h>
#include <net/udp.h>
#include <net/tcp.h>
#include "packet-filter/kernel/ip_tables.h"

struct in_device;
#include <net/route.h>
#include "packet-filter/extentions/ipt_LOG.h"

/* Make init and cleanup non-static, so gcc doesn't warn about unused,
   but don't export the symbols */
EXPORT_NO_SYMBOLS;

/* One level of recursion won't kill us */
static void dump_packet(struct iphdr *iph, unsigned int len, int recurse)
{
	void *protoh = (u_int32_t *)iph + iph->ihl;
	unsigned int datalen = len - iph->ihl * 4;

	/* Important fields:
	 * TOS, len, DF/MF, fragment offset, TTL, src, dst, options. */
	printk("SRC=%lu.%lu.%lu.%lu DST=%lu.%lu.%lu.%lu ",	
	       (ntohl(iph->saddr)>>24)&0xFF,
	       (ntohl(iph->saddr)>>16)&0xFF,
	       (ntohl(iph->saddr)>>8)&0xFF,
	       (ntohl(iph->saddr))&0xFF,
	       (ntohl(iph->daddr)>>24)&0xFF,
	       (ntohl(iph->daddr)>>16)&0xFF,
	       (ntohl(iph->daddr)>>8)&0xFF,
	       (ntohl(iph->daddr))&0xFF);

	printk("LEN=%u TOS=0x%02X TTL=%u ID=%u ",
	       ntohs(iph->tot_len), iph->tos, iph->ttl, ntohs(iph->id));

	if (ntohs(iph->frag_off) & IP_CE)
		printk("CE ");
	if (ntohs(iph->frag_off) & IP_DF)
		printk("DF ");
	if (ntohs(iph->frag_off) & IP_MF)
		printk("MF ");

	if (ntohs(iph->frag_off) & IP_OFFSET)
		printk("FRAG:%u ", ntohs(iph->frag_off) & IP_OFFSET);

	if (iph->ihl * 4 != sizeof(struct iphdr)) {
		unsigned int i;

		printk("OPT (");
		for (i = sizeof(struct iphdr); i < iph->ihl * 4; i++)
			printk("%02X", ((u_int8_t *)iph)[i]);
		printk(") ");
	}

	switch (iph->protocol) {
	case IPPROTO_TCP: {
		struct tcphdr *tcph = protoh;

		printk("PROTO=TCP ");

		if (ntohs(iph->frag_off) & IP_OFFSET)
			break;

		if (datalen < sizeof (*tcph)) {
			printk("INCOMPLETE [%u bytes] ", datalen);
			break;
		}

		printk("SPT=%u DPT=%u ",
		       ntohs(tcph->source), ntohs(tcph->dest));
		printk("SEQ=%lu ACK=%lu ",
		       ntohl(tcph->seq), ntohl(tcph->ack_seq));
		printk("WINDOW=%u ", ntohs(tcph->window));
		printk("RES=0x%06lx ", 
		       ntohl(tcph->res1) << 2 | ntohl(tcph->res2));
		if (tcph->urg)
			printk("URG ");
		if (tcph->ack)
			printk("ACK ");
		if (tcph->psh)
			printk("PSH ");
		if (tcph->rst)
			printk("RST ");
		if (tcph->syn)
			printk("SYN ");
		if (tcph->fin)
			printk("FIN ");
		printk("URGP=%u ", ntohs(tcph->urg_ptr));

		if (tcph->doff * 4 != sizeof(struct tcphdr)) {
			unsigned int i;

			printk("OPT (");
			for (i =sizeof(struct tcphdr); i < tcph->doff * 4; i++)
				printk("%02X", ((u_int8_t *)tcph)[i]);
			printk(") ");
		}
		break;
	}
	case IPPROTO_UDP: {
		struct udphdr *udph = protoh;

		printk("PROTO=UDP ");

		if (ntohs(iph->frag_off) & IP_OFFSET)
			break;

		if (datalen < sizeof (*udph))
			printk("INCOMPLETE [%u bytes] ", datalen);

		printk("SPT=%u DPT=%u LEN=%u ",
		       ntohs(udph->source), ntohs(udph->dest),
		       ntohs(udph->len));
		break;
	}
	case IPPROTO_ICMP: {
		struct icmphdr *icmph = protoh;
		static size_t required_len[NR_ICMP_TYPES+1]
			= { [ICMP_ECHOREPLY] = 4,
			    [ICMP_DEST_UNREACH] 
			    = 8 + sizeof(struct iphdr) + 8,
			    [ICMP_SOURCE_QUENCH]
			    = 8 + sizeof(struct iphdr) + 8,
			    [ICMP_REDIRECT]
			    = 8 + sizeof(struct iphdr) + 8,
			    [ICMP_ECHO] = 4,
			    [ICMP_TIME_EXCEEDED]
			    = 8 + sizeof(struct iphdr) + 8,
			    [ICMP_PARAMETERPROB]
			    = 8 + sizeof(struct iphdr) + 8,
			    [ICMP_TIMESTAMP] = 20,
			    [ICMP_TIMESTAMPREPLY] = 20,
			    [ICMP_ADDRESS] = 12,
			    [ICMP_ADDRESSREPLY] = 12 };

		printk("PROTO=ICMP ");

		if (ntohs(iph->frag_off) & IP_OFFSET)
			break;

		if (datalen < 4) {
			printk("INCOMPLETE [%u bytes] ", datalen);
			break;
		}

		printk("TYPE=%u CODE=%u ", icmph->type, icmph->code);

		if (icmph->type <= NR_ICMP_TYPES
		    && required_len[icmph->type]
		    && datalen < required_len[icmph->type]) {
			printk("INCOMPLETE [%u bytes] ", datalen);
			break;
		}

		switch (icmph->type) {
		case ICMP_ECHOREPLY:
		case ICMP_ECHO:
			printk("ID=%u SEQ=%u ", 
			       ntohs(icmph->un.echo.id), 
			       ntohs(icmph->un.echo.sequence));
			break;

		case ICMP_PARAMETERPROB:
			printk("PARAMETER=%lu ", 
			       ntohl(icmph->un.gateway) >> 24);
			/* Fall through */
		case ICMP_REDIRECT:
			if (icmph->type == ICMP_REDIRECT)
				printk("GATEWAY=%lu.%lu.%lu.%lu ", 
				       (ntohl(icmph->un.gateway)>>24)&0xFF,
				       (ntohl(icmph->un.gateway)>>16)&0xFF,
				       (ntohl(icmph->un.gateway)>>8)&0xFF,
				       (ntohl(icmph->un.gateway))&0xFF);
		case ICMP_DEST_UNREACH:
		case ICMP_SOURCE_QUENCH:
		case ICMP_TIME_EXCEEDED:
			if (recurse) {
				printk("[");
				dump_packet((struct iphdr *)(icmph + 1),
					    datalen - sizeof(struct iphdr), 0);
				printk("]");
			}

			if (icmph->type == ICMP_DEST_UNREACH
			    && icmph->code == ICMP_FRAG_NEEDED)
				printk("MTU=%u ", ntohs(icmph->un.frag.mtu));
		}
		break;
	}
	default:
		printk("PROTO %u ", iph->protocol);
	}
}

static int ipt_log_target(struct sk_buff *skb,
			  const char *indev,
			  const char *outdev,
			  const union ipt_targinfo *info,
			  unsigned int pos)
{
	struct iphdr *iph = skb->nh.iph;
	struct ipt_log_info *loginfo = (void *)info->other;
	char prefix[sizeof(KERN_WARNING) + sizeof(loginfo->prefix)];

	prefix[0] = '<';
	prefix[1] = '0' + (loginfo->level % 8);
	prefix[2] = '>';
	memcpy(prefix+3, loginfo->prefix, sizeof(loginfo->prefix));
	prefix[sizeof(prefix)-1] = '\0';

	printk("%sIN=%s OUT=%s ", prefix, indev, outdev);
	dump_packet(iph, skb->len, 1);
	printk("\n");

	return pos + 1;
}		

static int ipt_log_checkentry(const union ipt_targinfo *info,
			      unsigned int pos,
			      unsigned int finalpos)
{
	struct ipt_log_info *loginfo = (void *)info->other;

	if (loginfo->level >= 8
	    || loginfo->prefix[sizeof(loginfo->prefix)-1] != '\0')
		return 0;

	return 1;
}

static struct ipt_target ipt_log_reg
= { { NULL, NULL }, "LOG", ipt_log_target, ipt_log_checkentry, NULL, 
    &__this_module };

int __init init(void)
{
    if (ipt_register_target(&ipt_log_reg))
	return -EINVAL;

    return 0;
}

void __exit cleanup(void)
{
    ipt_unregister_target(&ipt_log_reg);
}

module_init(init);
module_exit(cleanup);
