/* Kernel module to match ICMP parameters. */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/icmp.h>

#include "packet-filter/extentions/ipt_icmp.h"
#include "packet-filter/kernel/ip_tables.h"

EXPORT_NO_SYMBOLS;
#ifdef DEBUG_CONNTRACK
#define duprintf(format, args...) printk(format , ## args)
#else
#define duprintf(format, args...)
#endif

/* Returns 1 if the type and code is matched by the range, 0 otherwise */
static inline int
type_code_match(u_int8_t test_type, u_int8_t min_code, u_int8_t max_code,
		u_int8_t type, u_int8_t code,
		int invert)
{
	return (type == test_type && code >= min_code && code <= max_code)
		^ invert;
}

static int
match(struct sk_buff *skb,
      const char *indev,
      const char *outdev,
      const union ipt_matchinfo *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{
	const struct icmphdr *icmp = hdr;

	if (offset == 0 && datalen < 2) {
		/* We've been asked to examine this packet, and we
		   can't.  Hence, no choice but to drop. */
		duprintf("Dropping evil ICMP tinygram.\n");
		*hotdrop = 1;
		return 0;
	}

	/* Must not be a fragment. */
	return !offset
		&& type_code_match(matchinfo->icmp.type,
				   matchinfo->icmp.code[0],
				   matchinfo->icmp.code[1],
				   icmp->type, icmp->code,
				   !!(matchinfo->icmp.invflags&IPT_ICMP_INV));
}

/* Called when user tries to insert an entry of this type. */
static int
checkentry(const struct ipt_ip *ip,
	   const union ipt_matchinfo *matchinfo)
{
	/* Must specify proto == ICMP, and no unknown invflags */
	return ip->proto == IPPROTO_ICMP
		&& !(matchinfo->icmp.invflags & ~IPT_ICMP_INV);
}

static struct ipt_match icmp_match
= { { NULL, NULL }, "icmp", &match, &checkentry, NULL, &__this_module };

int __init init(void)
{
	return ipt_register_match(&icmp_match);
}

void __exit cleanup(void)
{
	ipt_unregister_match(&icmp_match);
}

module_init(init);
module_exit(cleanup);
