/* Kernel module to match TCP parameters. */
#include <linux/module.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>

#include "packet-filter/extentions/ipt_tcp.h"
#include "packet-filter/kernel/ip_tables.h"

#ifdef DEBUG_CONNTRACK
#define duprintf(format, args...) printk(format , ## args)
#else
#define duprintf(format, args...)
#endif

EXPORT_NO_SYMBOLS;

/* Returns 1 if the port is matched by the range, 0 otherwise */
static inline int
port_match(u_int16_t min, u_int16_t max, u_int16_t port, int invert)
{
	return (port >= min && port <= max) ^ invert;
}

static int
find_option(u_int8_t option,
	    const struct tcphdr *tcp,
	    u_int16_t datalen,
	    int invert,
	    int *hotdrop)
{
	unsigned int i = sizeof(struct tcphdr);
	const u_int8_t *opt = (u_int8_t *)tcp;

	/* If we don't have the whole header, drop packet. */
	if (tcp->doff * 4 > datalen) {
		*hotdrop = 1;
		return 0;
	}

	while (i < tcp->doff * 4) {
		if (opt[i] == option) return !invert;
		if (opt[i] < 2) i++;
		else i += opt[1]?:1;
	}

	return invert;
}

static int
match(struct sk_buff *skb,
      const char *indev,
      const char *outdev,
      const union ipt_matchinfo *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{
	const struct tcphdr *tcp = hdr;

	/* To quote Alan:

	   Don't allow a fragment of TCP 8 bytes in. Nobody normal
	   causes this. Its a cracker trying to break in by doing a
	   flag overwrite to pass the direction checks.
	*/

	if (offset == 1) {
		duprintf("Dropping evil TCP offset=1 frag.\n");
		*hotdrop = 1;
		return 0;
	} else if (offset == 0 && datalen < sizeof(struct tcphdr)) {
		/* We've been asked to examine this packet, and we
		   can't.  Hence, no choice but to drop. */
		duprintf("Dropping evil TCP offset=0 tinygram.\n");
		*hotdrop = 1;
		return 0;
	}

	/* FIXME: Try tcp doff >> packet len against various stacks --RR */

#define FWINVTCP(bool,invflg) ((bool) ^ !!(matchinfo->tcp.invflags & invflg))

	/* Must not be a fragment. */
	return !offset
		&& port_match(matchinfo->tcp.spts[0], matchinfo->tcp.spts[1],
			      ntohs(tcp->source),
			      !!(matchinfo->tcp.invflags & IPT_TCP_INV_SRCPT))
		&& port_match(matchinfo->tcp.dpts[0], matchinfo->tcp.dpts[1],
			      ntohs(tcp->dest),
			      !!(matchinfo->tcp.invflags & IPT_TCP_INV_DSTPT))
		&& FWINVTCP((((unsigned char *)tcp)[13]
			     & matchinfo->tcp.flg_mask)
			    == matchinfo->tcp.flg_cmp,
			    IPT_TCP_INV_FLAGS)
		&& (!matchinfo->tcp.option
		    || find_option(matchinfo->tcp.option, tcp, datalen,
				   matchinfo->tcp.invflags
				   & IPT_TCP_INV_OPTION,
				   hotdrop));
}

/* Called when user tries to insert an entry of this type. */
static int
checkentry(const struct ipt_ip *ip,
	   const union ipt_matchinfo *matchinfo)
{
	/* Must specify proto == TCP, and no unknown invflags */
	return ip->proto == IPPROTO_TCP
		&& !(matchinfo->tcp.invflags & ~IPT_TCP_INV_MASK);
}

static struct ipt_match tcp_match
= { { NULL, NULL }, "tcp", &match, &checkentry, NULL, THIS_MODULE };

int __init init(void)
{
	return ipt_register_match(&tcp_match);
}

void __exit cleanup(void)
{
	ipt_unregister_match(&tcp_match);
}

module_init(init);
module_exit(cleanup);
