/* Shared library add-on to iptables to add LOG support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <getopt.h>
#include "packet-filter/userspace/iptables.h"
#include "packet-filter/kernel/ip_tables.h"
#include "packet-filter/extentions/ipt_LOG.h"

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"LOG v%s options:\n"
" --log-level level		Level of logging (numeric or see syslog.conf)\n"
" --log-prefix prefix		Prefix log messages with this prefix.\n\n",
NETFILTER_VERSION);
}

static struct option opts[] = {
	{ "log-level", 1, 0, '!' },
	{ "log-prefix", 1, 0, '#' },
	{ 0 }
};

static void
init(union ipt_targinfo *targinfo, unsigned int *nfcache)
{
	struct ipt_log_info *info = (struct ipt_log_info *)targinfo->other;

	info->level = LOG_WARNING;
	memset(info->prefix, 0, sizeof(info->prefix));
}

struct ipt_log_names {
	const char *name;
	unsigned int level;
};

static struct ipt_log_names ipt_log_names[] 
= { { "alert", LOG_ALERT },
    { "crit", LOG_CRIT },
    { "debug", LOG_DEBUG },
    { "emerg", LOG_EMERG },
    { "error", LOG_ERR },		/* DEPRECATED */
    { "info", LOG_INFO },
    { "notice", LOG_NOTICE },
    { "panic", LOG_EMERG },		/* DEPRECATED */
    { "warning", LOG_WARNING }
};

static u_int8_t
parse_level(const char *level)
{
	int lev;

	lev = string_to_number(level, 0, 7);
	if (lev == -1) {
		unsigned int i = 0;

		for (i = 0; 
		     i < sizeof(ipt_log_names) / sizeof(struct ipt_log_names);
		     i++) {
			if (strncasecmp(level, ipt_log_names[i].name,
					strlen(level)) == 0) {
				if (lev != -1)
					exit_error(PARAMETER_PROBLEM,
						   "log-level `%s' ambiguous",
						   level);
				lev = ipt_log_names[i].level;
			}
		}

		if (lev == -1)
			exit_error(PARAMETER_PROBLEM,
				   "log-level `%s' unknown", level);
	}

	return (u_int8_t)lev;
}

#define IPT_LOG_OPT_LEVEL 0x01
#define IPT_LOG_OPT_UNLIMIT 0x02
#define IPT_LOG_OPT_PREFIX 0x04

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      struct ipt_entry *entry)
{
	struct ipt_log_info *info = 
		(struct ipt_log_info *)entry->targinfo.other;

	switch (c) {
	case '!':
		if (*flags & IPT_LOG_OPT_LEVEL)
			exit_error(PARAMETER_PROBLEM, 
				   "Can't specify --log-level twice");
			
		if (check_inverse(optarg, &invert))
			exit_error(PARAMETER_PROBLEM, 
				   "Unexpected `!' after --log-level");

		info->level = parse_level(optarg);
		*flags |= IPT_LOG_OPT_LEVEL;
		break;

	case '#':
		if (*flags & IPT_LOG_OPT_PREFIX)
			exit_error(PARAMETER_PROBLEM, 
				   "Can't specify --log-prefix twice");
			
		if (check_inverse(optarg, &invert))
			exit_error(PARAMETER_PROBLEM, 
				   "Unexpected `!' after --log-prefix");

		if (strlen(optarg) > sizeof(info->prefix) + 1)
			exit_error(PARAMETER_PROBLEM,
				   "Maximum prefix length %u for --log-prefix",
				   sizeof(info->prefix) - 1);

		strcpy(info->prefix, optarg);
		*flags |= IPT_LOG_OPT_PREFIX;
		break;

	default:
		return 0;
	}

	return 1;
}

/* Final check; nothing. */
static void final_check(unsigned int flags)
{
}

/* Prints out the union ipt_targinfo. */
static void print(const union ipt_targinfo *info, int numeric)
{
	unsigned int i = 0;
	struct ipt_log_info *loginfo = (struct ipt_log_info *)info->other;

	printf("LOG ");
	if (numeric)
		printf("level %u ", loginfo->level);
	else {
		for (i = 0; 
		     i < sizeof(ipt_log_names) / sizeof(struct ipt_log_names);
		     i++) {
			if (loginfo->level == ipt_log_names[i].level) {
				printf("level %s ", ipt_log_names[i].name);
				break;
			}
		}
		if (i == sizeof(ipt_log_names) / sizeof(struct ipt_log_names))
			printf("UNKNOWN level %u ", loginfo->level);
	}

	if (strcmp(loginfo->prefix, "") != 0)
		printf("prefix `%s' ", loginfo->prefix);
}

struct iptables_target log
= { NULL,
    "LOG",
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    opts
};

void _init(void)
{
	register_target(&log);
}
