/* Support routines for when iptables is in userspace */
#include "ip_tables_usersupp.h"
#include "fake_kernel_interface.h"
#include <stdarg.h>
#include <stdlib.h>

#define ROUND_UP(s,n) (((s)+(n)-1)/(n)*(n))

void panic(const char message[], ...)
{
	va_list args;
	va_start(args, message);

	vfprintf(stderr, message, args);
	va_end(args);

	abort();
}

void *kmalloc(size_t size, int pri)
{
	void **ptr = malloc(ROUND_UP(size, sizeof(void *)) + sizeof(void *)*2);
	ptr[0] = (void *)size;
	memset((unsigned char *)(ptr+1) + size, 0xde, sizeof(void *));
	return ptr + 1;
}

void kfree(void *p)
{
	void **ptr = p;
	size_t size = (size_t)ptr[-1];
	unsigned char *end = (unsigned char *)p + size;
	unsigned int i;

	for (i = 0; i < sizeof(void *); i++) {
		if (end[i] != 0xde)
			panic("Problem with freeing %p, size %u\n",
			      p, (unsigned int)size);
	}

	free(p-1);
}

void request_module(const char modulename[])
{
	printk("Request for module `%s'.\n", modulename);
}

int copy_from_user(void *to, void *from, size_t size)
{
	memcpy(to, from, size);
	return 0;
}

int copy_to_user(void *to, void *from, size_t size)
{
	memcpy(to, from, size);
	return 0;
}

#include "ip_tables_usersupp.h"
#include "ip_tables.h"
#include <linux/netfilter_dev.h>
#include <ctype.h>
#include <assert.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/poll.h>
#include <sys/ioctl.h>

static const char *devnames[] =
{
	"/dev/netfilter_ip_raw_input",
	"/dev/netfilter_ip_pre_routing",
	"/dev/netfilter_ip_input",
	"/dev/netfilter_ip_forward",
	"/dev/netfilter_ip_output",
	"/dev/netfilter_ip_outgoing",
};

#define GET_NAME(name, hooknum) 			\
do { 							\
	char *p; 					\
	strcpy(name, "/dev/netfilter_ip_"); 		\
	strcat(name, #hooknum + sizeof("NF_IP_") - 1); 	\
	for (p = name; *p; p++) *p = tolower(*p);	\
} while(0)

static int devhandle[NF_IP_NUMHOOKS];
static struct nf_hook_ops *hooks[NF_IP_NUMHOOKS];
struct pollfd polls[NF_IP_NUMHOOKS+1];
static int num_regs = 0;

int nf_register_hook(struct nf_hook_ops *hook)
{
	/* Here's where we do the real work. */
	assert(hook->pf == PF_INET);
	assert(hook->hooknum < NF_IP_NUMHOOKS);

	if (devhandle[hook->hooknum] < 0) {
		char command[80];

		devhandle[hook->hooknum]
			= open(devnames[hook->hooknum], O_RDWR);
		if (devhandle[hook->hooknum] < 0) {
			fprintf(stderr, "Can't open `%s': %s\n",
				devnames[hook->hooknum], strerror(errno));
			return -errno;
		}

		/* Crude, but it'll do for the moment. */
		sprintf(command, "insmod ./ip_usermod.o hook=%i -o ip_usermod%i",
			hook->hooknum, hook->hooknum);
		if (system(command) != 0) {
#if 0
			fprintf(stderr, "Problem running `%s'.\n", command);
			return -1;
#endif
		}
		polls[num_regs].fd = devhandle[hook->hooknum];
		polls[num_regs].events = POLLIN;
		num_regs++;
	}
	hook->next = hooks[hook->hooknum];
	hooks[hook->hooknum] = hook;

	return 0;
}

static unsigned int find_hook(int fd)
{
	unsigned int i;

	for (i = 0; i < NF_IP_NUMHOOKS; i++)
		if (devhandle[i] == fd) return i;

	panic("No handle matching fd=%i.\n", fd);
}

static void call_hook(struct nf_hook_ops *hook, int fd)
{
	struct nfdev_head head;
	struct sk_buff *skb;
	struct nfdev_ioctl ioc;

	if (read(fd, &head, sizeof(head)) != sizeof(head))
		panic("Failure to read header from fd %i.\n", fd);

	skb = malloc(head.pktlen + sizeof(struct sk_buff));
	skb->nfmark = head.nfmark;
	skb->nh.iph = (struct iphdr *)((char *)skb + sizeof(struct sk_buff));
	if (read(fd, skb->nh.iph, head.pktlen) != head.pktlen)
		panic("Failure to read packet len %u from fd %i.\n",
		      head.pktlen, fd);

	ioc = ((struct nfdev_ioctl) { 0, head.pktlen, head.nfmark });
	switch (hook->hook(hook->hooknum, &skb, head.iniface, head.outiface)) {
	case NF_ACCEPT:
		ioc.verdict = NF_ACCEPT;
		break;

	case NF_QUEUE:
		ioc.verdict = NF_DROP;
		break;

	case NF_DROP:
		ioc.verdict = NF_DROP;
		break;
		
	default:
		panic("Wierd verdict for hook %u\n", hook->hooknum);
	}

	if (ioctl(fd, NFDIOSVERDICT, &ioc) != 0)
		panic("ioctl: %s\n", strerror(errno));
}

static void call_hooks(unsigned int hook)
{
	struct nf_hook_ops *list;

	for (list = hooks[hook]; list; list = list->next)
		call_hook(list, devhandle[hook]);
}

int main()
{
	unsigned int i;

	for (i = 0; i < NF_IP_NUMHOOKS; i++) devhandle[i] = -1;

	/* Call init function */
	ip_fw_init();

	/* Receive commands from fifo. */
	if (mknod(FIFONAME, 0600, S_IFIFO) < 0 && errno != EEXIST)
		panic("mknod %s: %s\n", FIFONAME, strerror(errno));

	polls[num_regs].fd = open(FIFONAME, O_RDWR);
	if (polls[num_regs].fd < 0)
		panic("opening %s: %s\n", FIFONAME, strerror(errno));
	polls[num_regs].events = POLLIN;

	/* Wait for packets. */
	while (poll(polls, num_regs+1, -1) > 0) {
		for (i = 0; i < num_regs; i++) {
			if (polls[i].revents & POLLIN)
				call_hooks(find_hook(polls[i].fd));
		}
		if (polls[num_regs].revents & POLLIN)
			read_command(polls[num_regs].fd);
	}

	perror("poll");
	return 0;
}
